#ifdef USE_AUDIO
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#ifdef USE_SDL1
#include <SDL/SDL.h>
#endif
#ifdef USE_SDL2
#include <SDL2/SDL.h>
#endif

#ifdef USE_NLLIBC
#include <nllibc.h>
#endif

#include "config.h"
#include "const.h"
#include "nlllib.h"
#include "audio.h"
#include "sdl.h"

#define SINNUM 64

static int sin_val[SINNUM + 1] = {
  0,     245,  491,  736,  980, 1224, 1467, 1710,
  1951, 2191, 2430, 2667, 2903, 3137, 3369, 3599,
  3827, 4052, 4276, 4496, 4714, 4929, 5141, 5350,
  5556, 5758, 5957, 6152, 6344, 6532, 6716, 6895,
  7071, 7242, 7410, 7572, 7730, 7883, 8032, 8176,
  8315, 8449, 8577, 8701, 8819, 8932, 9040, 9142,
  9239, 9330, 9415, 9495, 9569, 9638, 9700, 9757,
  9808, 9853, 9892, 9925, 9952, 9973, 9988, 9997, 10000
};

static int int_sin(int deg)
{
  while (deg <           0) deg += (SINNUM*4);
  while (deg >= (SINNUM*4)) deg -= (SINNUM*4);
  if (deg > (SINNUM*2))
    return -int_sin(deg - (SINNUM*2));
  if (deg > SINNUM)
    return int_sin((SINNUM*2) - deg);
  return sin_val[deg];
}

#define SCALENUM 12

static int freq[SCALENUM] = {
  16744036, /* C9  */
  17739688, /* C#9 */
  18794545, /* D9  */
  19912127, /* D#9 */
  21096164, /* E9  */
  22350607, /* F9  */
  23679643, /* F#9 */
  25087708, /* G9  */
  26579501, /* G#9 */
  28160000, /* A9  */
  29834481, /* A#9 */
  31608531, /* B9  */
};

struct audio_data {
  int freq;
  int amp;
  int step;
  int samplefreq;
};

static int init(void)
{
  static int initialized = 0;

  if (!initialized) {
#ifdef USE_SDL
    sdl_start();
#endif
    initialized = 1;
  }

  return 0;
}

#ifdef USE_SDL
static void sdl_callback(void *userdata, unsigned char *stream, int len)
{
  short *frames;
  int i, n, a;
  struct audio_data *audiodata;

  audiodata = (struct audio_data *)userdata;

  frames = (short *)stream;
  n = len / 2;

  for (i = 0; i < n; i++) {
    a = int_sin(audiodata->step * audiodata->freq * (SINNUM*4) / audiodata->samplefreq);
    frames[i] = a * audiodata->amp / 10;
    audiodata->step++;
  }
}
#endif

int audio_sine(int scale, int octave, int msec)
{
  int f, i;
  struct audio_data audiodata;
#ifdef USE_SDL
  SDL_AudioSpec spec;
#endif

  init();

  f = freq[scale];
  for (i = octave; i < 9; i++)
    f >>= 1;
  f /= 1000;

#define SAMPLING_FREQ (256 * (128 + 32 + 16))
  audiodata.freq = f;
  audiodata.amp  = 32;
  audiodata.step = 0;
  audiodata.samplefreq = SAMPLING_FREQ;

#ifdef USE_SDL
  memset(&spec, 0, sizeof(spec));

  spec.freq = SAMPLING_FREQ;
  spec.format = AUDIO_S16;
  spec.channels = 1;
  spec.samples = 256;
  spec.callback = sdl_callback;
  spec.userdata = &audiodata;

  SDL_OpenAudio(&spec, NULL);
  SDL_PauseAudio(0);
  SDL_Delay(msec);
  SDL_PauseAudio(1);
#endif

  return 0;
}

static int code2scale[] = { 0, 2, 4, 5, 7, 9, 11 };

int audio_play(const char *string)
{
  int scale, octave, interval;
  int r = 0;
  const char *s, *p, *upper, *lower;

  init();

  upper = "CDEFGAB";
  lower = "cdefgab";

  for (s = string; *s;) {
    if ((p = strchr(upper, *s)) == NULL) {
      if ((p = strchr(lower, *s)) == NULL) {
	return NLL_ERRCODE_VALUE_INVALID_TYPE;
      } else {
	scale = code2scale[p - lower];
      }
    } else {
      scale = code2scale[p - upper];
    }
    s++;

    if (*s && (*s == '#')) {
      scale++;
      s++;
    }

    octave = 3;
    if (*s && isdigit(*s)) {
      octave = *s - '0';
      s++;
    }

    while (scale >= SCALENUM) {
      scale -= SCALENUM;
      octave++;
    }

    interval = 1;
    if (*s && isdigit(*s)) {
      interval = *s - '0';
      s++;
    }

    if ((r = audio_sine(scale, octave, interval * 250)) < 0)
      break;
  }

  return r;
}

#endif
